library hakkit.hakk.builder.basic;

import 'dart:math';
import 'dart:io' show Platform;

part 'src/basic/block.dart';
part 'src/basic/flow_control.dart';
part 'src/basic/predefines.dart';
part 'src/basic/_operator_class.dart';
part 'src/basic/_hakk_object.dart';
part 'src/basic/_expression_node.dart';
part 'src/basic/_code_block.dart';

final List<int> _ifStack = new List<int>();
final List<int> _whileStack = new List<int>();
final List _expressionQueue = new List();
const int _logicMemIndexerOffset = 40;
const int _functionIndexerOffset = 80;
int _memIndexer = 0;
int _functionIndexer = _functionIndexerOffset;
int _logicMemIndexer = _logicMemIndexerOffset;
int _logicLabelIndexer = 0;
_CodeBlock _currentCodeBlock;
String _hakkName = '';
final List<_BlockDeclaration> _blockDeclarations = new List<_BlockDeclaration>();
final Map<String, _CodeBlock> _codeBlocks = new Map<String, _CodeBlock>();
final String _CR = Platform.isWindows ? '\r\n' : '\r';

/// Declares the beginning of a new hakk program. Note that this will erase
/// any previous declarations that were not already generated with [build].
/// 
/// ## Example
///     hakk('my_hakk');
///     //... rest of hakk code here ...
///     print(build());
/// 
/// @param name is the name of the hakk.
void hakk(String name){
  _init();
  _hakkName = name;
  _blockDeclarations.add(new _Core());
}

/// Defines a value type for use in expressions.
/// 
/// Supported value types are: [String], [num], and [bool].
/// 
/// This is typically only needed in expressions where the value type is on the
/// left-hand side of an expression operator.  The reason for this is related to the
/// way that Dart handles operator overrides.
/// 
/// ## Example
///      hakk('val_example');
///      Block.system(0, 1);
///      
///      // val() is not needed for simple use cases such as:
///      output('a string');
///      output(42);
///      output(true);
///      
///      // but if you want to use literals in an expression, then
///      // you will need to wrap them in val() if they are on the
///      // left-hand side of an operation.  Determining The "side" 
///      // can be a bit tricky in complex expressions.  It's safest
///      // then to wrap all literals in val()
///      
///      // this works
///      output(core0 + 10);
///      
///      // this doesn't work
///      output(10 + core0);
///      // use instead
///      output(val(10) + core0);
///      
///      // if the expression is complex, then wrap them to be safe
///      output(val(10) + (core1 * val(4)));
///      
///      // same goes for conditional expressions
///      // this outputs "true":
///      output(val(10) < val(20));
/// 
/// @param value Must either be of type [String], [num], or [bool].
_HakkObject val(value){
  if (value is! String && value is! num && value is! bool){
    throw 'A HakkObject value type must be either a string, number, or boolean.';
  }
  return new _HakkObject.literal(value);
}

/// Defines a memory reference to a location in a [Block.memory] block.
/// 
/// In addition to the 16 pre-defined core memory registers, available by default
/// to any hakk.  It is possible to expand the available memory space by adding a
/// [Block.memory] block to a hakk, and then using this mem function to store and
/// retrieve information.  
/// 
/// Each memory block provides an additional 64 memory locations.
/// 
/// ## Example
///     hakk('mem_example');
///     Block.system(0, 1);
///     
///     // first define a memory block for the hakk.
///     Block.memory('m1', 0, -1);
///  
///     core0 << 5;
///     
///     // here we store the value in core0 into position 0 of the memory block.
///     mem('m1', 0) << core0;
///   
///     // you can also define concrete representions of memory block locations.
///     // Note that these do not necessarily have to be defined in the hakk code
///     // itself.  You could define them ahead of time and use them in multiple
///     // hakks (assuming the name of the memory block is the same). 
///     final mem0 = mem('m1', 0);
///     final mem1 = mem('m1', 1);
///   
///     // Once defined they can be used in writing code, just like the 
///     // pre-defined core0-core15 memory registers.
///     mem1 << mem0 * 3;
///   
///     output(mem1);
///   
///     print(build());
///   
/// @param memoryBlockName is the name of the [Block.memory] block as you 
/// defined it.
/// @param index is the location in the memory block where you want to store or
/// retrieve information.
/// @returns An object suitable for use in code.
_HakkObject mem(String memoryBlockName, int index){
  return new _HakkObject.reference(memoryBlockName, index);
}


/// Takes all previous declarations since the last [hakk] declaration, and 
/// generates hakkML source code.
/// 
/// ## Example
///     hakk('mem_example');
///     Block.system(0, 1);
///     
///     output('Hello World!');
///     
///     // print the compiled hakkML to the console.
///     print(build());
/// 
/// @returns a String that contains the generated hakkML source code.
String build(){
  if (_ifStack.isNotEmpty){
    throw 'IF/ENDIF statements are not balanced.';
  }
  
  if (_whileStack.isNotEmpty){
    throw 'WHILE/ENDWHILE statements are not balanced.';
  }
  
  if (!_blockDeclarations.any((b) => b.name == '__sys__')){
    throw 'A required system block (Block.system] declaration was not found.';
  }
  
  final sb = new StringBuffer();
  sb.write('name: "${_hakkName}"$_CR');
  sb.write('blocks:$_CR');
  _blockDeclarations.forEach((b) => sb.write(b.toString()));
  sb.write('code:$_CR');
  _codeBlocks.forEach((name, code){
    sb.write('  ${name}:$_CR');
    code.lines.forEach((line){
      sb.write('    - ${line}$_CR');
    });
  });
  return sb.toString();
}

/// Explicity declares the entry code block.  This is done for you
/// (implicitly) if you start writing code after declaring blocks.
/// 
/// This declaration is typically used when you want to the entry
/// code after some other [Block.instruction] block, instead of before.
/// 
/// ## Example without (implicitly defined for you)
///     hakk('hello_world');
///     Block.system(0, 1);
///  
///     output("Hello World!");
///
///     print(build());
///     
/// ## Example with (explicitly defined by you)
///     hakk('hello_world');
///     Block.system(0, 1);
///  
///     codeEntry();
///       output("Hello World!");
///
///     print(build());
///     
/// ## Typical use example (entry code not declared first)
///     hakk('hello_world');
///     Block.system(0, 1);
///     Block.instruction('other_code', 0, -1);
///     
///     // declaring code for this instruction block first
///     codeBlock('other_code');
///       output('in other code');
///       EXIT(0);
///        
///     codeEntry();
///       output("Hello World!");
///       GOTO('other_code');
///
///     print(build());
void codeEntry() => codeBlock('core');


/// Declares the beginning of a new [Block.instruction] code block.  Any code 
/// written after this declaration will be associated to that block.
/// 
/// ## Example
///     hakk('hello_world');
///     Block.system(0, 1);
///     
///     // define a instruction block for the code
///     Block.instruction('other_code', 0, -1);
///     
///     // this code will belong to the default entry (core) block.
///     output('Hello World!');
///     GOTO('other_code');
///     
///     // declaring code for the instruction block
///     codeBlock('other_code');
///       // this code belongs to the instruction blocked named 'other_code'
///       output('in other code');
///       EXIT(0);
///
///     print(build());
///  
/// @param name is the name of the [Block.instruction] to associate the code to.  If
/// the names don't match, the hakk will get a runtime error.
void codeBlock(String name){
  if (_codeBlocks.containsKey(name)){
    throw 'There is already a code block declared with the name "${name}".';
  }
  
  _codeBlocks[name] = new _CodeBlock();
  _currentCodeBlock = _codeBlocks[name];
  _memIndexer = 0;
}

void _init(){
  _blockDeclarations.clear();
  _expressionQueue.clear();
  _ifStack.clear();
  _whileStack.clear();
  _codeBlocks.clear();
  _currentCodeBlock = null;
  _memIndexer = 0;
  _functionIndexer = _functionIndexerOffset;
  _logicMemIndexer = _logicMemIndexerOffset;
  _logicLabelIndexer = 0;
}

_wrapLiteral(value){
  if (value is String || value is num || value is bool){
    return val(value);
  }else if (value is _HakkObject){
    return value;
  }else if (value is _ExpressionNode) {
    return value;
  }else{
    throw 'Type not supported: ${value.runtimeType}';
  }
}





