part of hakkit.hakk.builder.basic;

/// Declares a condition loop that continues while the given [expr] evaluates to
/// true.
/// 
/// All WHILE() statements just eventually end with a [ENDWHILE] statement.
/// 
/// ## Limitations
/// Complex looping scenarios are not well supported.  Use care when nesting WHILE
/// loops, as this will likely yield unexpected results.  You may need to use the
/// advanced API if you require more complex looping implementations, or [inline]
/// your own hakkML directly.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///          
///     // loop that ends when core0 >= 10
///     WHILE(core0 < 10);
///       output(core0);
///       inc(core0);
///     ENDWHILE();
///     
///     // infinite loop.  Very common in a hakk to have a main loop that
///     // is continuous.
///     WHILE(true);
///       output('to infinity and beyond!');
///     ENDWHILE();
/// 
/// # Related
/// * [ENDWHILE]
void WHILE(expr){
  expr = _wrapLiteral(expr);
  
  _whileStack.add(++_logicLabelIndexer);
  _HakkObject.addLine('lbl: __while_test_${_whileStack.last}__');

  IF(expr);
}

/// Defines the end of a [WHILE] block.  When the [WHILE] expression evaluates to
/// false, code immediately following ENDWHILE will execute.
/// 
/// # Related
/// * [WHILE]
void ENDWHILE(){
  if (_whileStack.isEmpty){
    throw 'Found ENDWHILE() without a corresponding WHILE(). Check to make sure your WHILE/ENDWHILE declarations are balanced.';
  }
  _HakkObject.addLine('goto: __while_test_${_whileStack.last}__');
  ENDIF();
  _whileStack.removeLast();
}

/// Provides a conditional mechanism that executes code inside the IF/ENDIF block 
/// when the given [expr] evaluates to true.
/// 
/// Any IF statement must have a corresponding [ENDIF].
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     core0 << 10
///     IF(core0 < 100);
///       output('core0 is less than 100');
///     ENDIF();
///     
/// # Related
/// * [ENDIF]
void IF(expr){
  expr = _wrapLiteral(expr);
  _ifStack.add(++_logicLabelIndexer);
 
  final i = _ifStack.length;
  final li = _logicLabelIndexer++;

  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('je: [__sys__.${expr.id}, true, __iftrue_${_ifStack.last}__]');
  }else{
    _HakkObject.addLine('set: [__sys__.${_logicMemIndexer}, ${expr}]');
    _HakkObject.addLine('je: [__sys__.${_logicMemIndexer}, true, __iftrue_${_ifStack.last}__]');    
  }
  _HakkObject.addLine('goto: __endif_${_ifStack.last}__');
  _HakkObject.addLine('lbl: __iftrue_${_ifStack.last}__');
  _logicMemIndexer = _logicMemIndexerOffset;
}

/// Defines the end of a [IF] block.
/// 
/// # Related
/// * [IF]
void ENDIF(){
  if (_ifStack.isEmpty){
    throw 'Found ENDIF() without a corresponding IF(). Check to make sure your IF/ENDIF declarations are balanced.';
  }
  _HakkObject.addLine('lbl: __endif_${_ifStack.last}__');
  _ifStack.removeLast();
}

/// Jumps code execution to the given [label].  A label can be a local [label] in
/// the current instruction block context, or it can be the name of another 
/// [Block.instruction], allowing transfer of control between instruction blocks.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     GOTO('this_place');
///     output('This will never be seen.');
///     
///     label('this_place');
///     output('Hi there.');
///     
/// @param label is the name of the label or instruction block to jump to.
void GOTO(String label){
  _HakkObject.addLine('goto: ${label}');
}

/// Initiates a call to a [SUB] subroutine, which is eventuall [RETURN]ed to the
/// instruction immediately following the GOSUB statement.
/// 
/// # parameter passing and value returns
/// The GOSUB/SUB/RETURN sequence also supports a rudimentary parameter passing
/// and value return.
/// 
/// ## Subroutines are NOT closures.
/// All values stored in memory blocks (such as [core0] or any memory block added
/// to a hakk) are essential "global".  This means that you must plan carefully
/// the use of these memory locations to prevent "clobbering", where one operation
/// overwrites a memory location that was not intended.
/// 
/// ## param/argument matching
/// It is important that when using GOSUB the number of parameter values passed, must
/// match the number of arguements expected to be received by the [SUB]routine. 
///  
/// # Always returns a value
/// [SUB]routines always return a value, even if one is not provided.  The default
/// return value is 0, if none is provided.  A GOSUB should always be assigned to some
/// memory location.  It will not execute correctly if called alone.
/// 
/// ## This works
///     core0 << GOSUB('foo');
/// 
/// ## This currently doesn't work
///     GOSUB('foo');
///     
/// ## This works as a workaround
///     GOSUB('foo').compile();
///  
/// # Limitations
/// Although GOSUB can technically participate in expressions, the use of GOSUB in
/// complex/compound expressions is not well supported, and may introduce unexpected
/// results.  Two GOSUBs used in the same compound expression may clobber each other's
/// temporary storage locations.
/// 
/// ## This is preferable
///     core0 << GOSUB('foo');
///     core1 << GOSUB('bar');
///     core2 << core0 + core1;
///     
/// ## This may give unexpected results.
///     core0 << GOSUB('foo') + GOSUB('bar');
/// 
/// # Example
///     hakk('GOSUB_example');
///     Block.system(0, 1);
///     
///     // GOSUB to a subroutine with no parameters
///     core0 << GOSUB('say_hi');
///     
///     // Pass 2 parameters to 'adder' and store the
///     // returned result in core0.
///     core0 << GOSUB('adder', [5, 5]);
///     // "10"
///     output(core0);
///     
///     SUB('say_hi');
///       output('hi');
///     RETURN();
///     
///     SUB('adder', [core3, core4]);
///       core5 << core3 + core4;
///     RETURN(core5);
///     
///     output('Hi there.');
///     
/// # Related
/// * [SUB]
/// * [RETURN]
/// 
/// param subroutineName is the name of the [SUB]routine to call.
/// param params optionally passes parameters to the [SUB]routine.
_ExpressionNode GOSUB(String subroutineName, [List params]){
  final node = new _ExpressionNode.call(subroutineName, params);
  _expressionQueue.add(node);
  return node;
}

/// Declares a subroutine block that can take 0-n arguments, run some code,
/// and then return a result back to the [GOSUB] that called it.
/// 
/// # Limitations
/// A Subroutine is NOT a closure.  All values and references used within it are
/// global to the hakk runtime context.
/// 
/// Subroutines do no support recursion (calling itself), nesting (a subroutine 
/// within another subroutine), or chaining (one subroutine calling another).  
/// Attempts to use them any of these ways will create unexpected results.  
/// 
/// Providing support for these features would requires heap and stack memory that
/// is outside the scope of what a hakk is intended for (specialized use in hakk
/// arenas).  
/// 
/// It is possible to emulate these behaviors yourself, by using the advanced API, or
/// by [inline]ing your own hakkML.  Proceed with caution.
/// 
/// # Example
/// See the [GOSUB] documentation for examples of how SUB is used.
/// 
/// # Related
/// * [GOSUB]
/// * [RETURN]
/// 
/// param subroutineName is the declared name of the subroutine that [GOSUB] will use
/// to call it.
/// param arguments is a list of memory locations that will store the values passed
/// by the [GOSUB] call.  Make sure the number of arguments matches the number of
/// parameters!
void SUB(String subroutineName, [List<_HakkObject> arguments]){
  if (_subroutineContext != null){
    throw 'Nest subroutines are not allowed.';
  }
  _subroutineContext = '${subroutineName}_protect';
  GOTO(_subroutineContext);
  label(subroutineName);
  if (arguments != null){
    arguments.reversed.forEach((a){
      if (a is! _HakkObject || (a is _HakkObject && a.isLiteral)){
        throw 'Argument must be a memory reference.';
      }
      a << pop();
    });
  }
}

/// Returns control from a [SUB]routine back to the [GOSUB] that called it.
/// 
/// RETURN may pass an optiona expression back to the GOSUB.
///
/// # Example
/// See the [GOSUB] documentation for examples of how RETURN is used.
///
/// # Related
/// * [SUB]
/// * [GOSUB]
void RETURN([expr]){
  if (_subroutineContext == null){
    throw 'Cannot have a subroutine [ret]urn without first declaring a subroutine().';
  }
  if (expr == null){
    _HakkObject.addLine('push: 0');
  }else{
    expr = _wrapLiteral(expr);
    _setClobber = true;
    push(expr);
  }
  
  _HakkObject.addLine('ret:');
  label(_subroutineContext);
  _subroutineContext = null;
}


/// Defines a label in code that can be used by [GOTO] to jump to a different
/// location in code.
/// 
/// # Example
/// See the [GOTO] documenation for examples of how label can be used.
/// 
/// # Related
/// * [GOTO]
/// 
/// @param label is the name which will be used by a [GOTO] statement to jump to this
/// location in the hakk.
void label(String label){
  _HakkObject.addLine('lbl: ${label}');
}


/// Unconditionally terminates processing of hakk code in the hakk.it virtual
/// machine.
/// 
/// This instruction is generally used in testing only.  It probably wouldn't help
/// your chances to win in the arenas if your hakk stops running!
/// 
/// @param expr is an expression that should result in a termination code for the
/// hakk.  At this time, this information is not exposed to the player, and can only
/// be inspected by internal tooling.
void EXIT(expr){
  expr = _wrapLiteral(expr);
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('exit: __sys__.${expr.id}');
  }else{
    _HakkObject.addLine('exit: ${expr}');    
  }
}