part of hakkit.hakk.builder.basic;


bool _setClobber = false;
String _subroutineContext;

/// Causes execution to halt for [expr] number of ticks in the hakk.it virtual
/// machine.
/// 
/// # Example
///     hakk('wait_example');
///     Block.system(0, 1);
///     
///     output('hi');
///     
///     // wait 60 ticks (about 1 second)
///     wait(60);
///     
///     output('there');
///     
///     core0 << 5;
///     core1 << 25;
///     
///     // wait supports expressions...
///     wait(core0 + core1);
///     output('finished.');
void wait(expr){
  expr = _wrapLiteral(expr);
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('wait: __sys__.${expr.id}');
  }else{
    _HakkObject.addLine('wait: ${expr}'); 
  }
}

/// Outputs [expr] to the STDOUT (usually the console).
/// 
/// This instruction doesn't really help in the arena, but is sometimes useful
/// for testing and demos.  Use sparingly in hakks intended for arena combat, as
/// output will waste precious ticks.
/// 
/// # Example
///     hakk('output_example');
///     Block.system(0, 1);
///     
///     // output a string
///     output('Hello World!');
///     
///     // output a numeric literal
///     output(42);
///     
///     // output a boolean
///     output(true);
///     
///     // output an expression
///     core0 << 5;
///     core1 << 15;
///     output(core0 + core1);
///     // "15"
void output(expr){
  expr = _wrapLiteral(expr);
  if (expr is _HakkObject){
    if (expr.isLiteral){
      _HakkObject.addLine('out: "${expr}"');
    }else{
      _HakkObject.addLine('out: ${expr}');
    }
  }else if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('out: __sys__.${expr.id}');
  }else{
    throw 'Invalid type passed to output.  This is probably a bug on our end. Sorry!';
  }
}

/// Increments a value by 1;
/// 
/// Equivalent to (using [core0] as an example):
/// 
///      core0 << core0 + 1;
///      
/// # Related
/// * [dec]
/// 
/// @param expr is a memory location with a value to be incremented.
void inc(expr){
  expr = _wrapLiteral(expr);
  if (expr.isLiteral){
    throw "The inc function only works on values in a memory reference.  Supply a memory reference instead.";
  }
  expr <<  expr + 1;
}

/// Decrements a value by 1;
/// 
/// Equivalent to (using [core0] as an example):
/// 
///      core0 << core0 - 1;
///      
/// # Related
/// * [inc]
/// 
/// @param expr is a memory location with a value to be incremented.
void dec(expr){
  expr = _wrapLiteral(expr);
  if (expr.isLiteral){
    throw "The inc function only works on values in a memory reference.  Supply a memory reference instead.";
  }
  expr << expr - 1;
}

/// Inlines hakkML into the source code.
/// 
/// ## Use With Caution
/// This is an advanced feature, and should not be used unless you have a really good
/// understanding of how the basic API builder works, along with a strong working
/// knowledge of hakkML.  The builder makes certain assumptions about the structure 
/// of code it generates, and inlining hakkML may break this.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     core0 << 42;
///     // outputs "42";
///     output(core0);
///     
///     //change core0 with inline
///     inline('set: [core.0, 24]');
///     
///     // outputs "24";
///     output(core0);
///     
///     // print the compiled hakkML to the console.
///     print(build());
///     
/// @param instructions If a string, is assumed to be a single instruction.  If a 
/// list of Strings, then is assumed to be a list of instructions.
void inline(instructions){
  if (instructions is String){
    _HakkObject.addLine(instructions);
  }else if (instructions is List){
    instructions.forEach((i) => _HakkObject.addLine(i));
  }
}

/// Generates a random integer between 0 (inclusive) and [expr] (exclusive)
/// 
/// # Example
///     hakk('random_example');
///     Block.system(0, 1);
///     
///     core0 << random(10);
///     output('Will output an integer between 0 - 9');
///     output(core0);
/// 
/// @param expr is a literal number or memory location of a number that represents
/// the upper bound of the random result.
_ExpressionNode random(expr){
  expr = _wrapLiteral(expr);
  final node = new _ExpressionNode(expr, expr, 'rnd');
  _expressionQueue.add(node);
  return node;
}

/// Pushes a value from [expr] onto the stack.
/// 
/// # The Stack
/// Each hakk comes with it's own value stack, allowing coders to utilize
/// techniques to pass information from one part of a hakk to another, without the
/// receiver having knowledge of "where" the information is.  Rather, the receiver
/// only needs to know how much information to [pop] off the stack.  Even the "amount"
/// can be added to the stack, provided a dynamic way to pass information around.
/// 
/// # The importance of balance
/// It's important that values pushed onto the stack are always [pop]ped from the
/// stack in balance (if you push 3 values, you should always pop 3 values elsewhere).
/// Failure to do this will likely result in stack overflow or stack underflow.
/// 
/// # Example
///     hakk('stack_example');
///     Block.system(0, 1);
///     
///     // push 3 values onto the stack, the last value being the "amount" of
///     // other values pushed.
///     core0 << 42;
///     // core0 can safely be used again because it's value is now pushed onto the
///     // stack.
///     push(core0);
///     push(-42);
///     push(2);
///     
///     // ..somewhere else in code
///     
///     core0 << pop();
///     output('how many items to retrieve:');
///     output(core0);
///     core1 << 0;
///
///     output('');
///     output('Items:');
///     WHILE(core1 <= core0);
///       core3 << pop();
///       output(core3);
///       inc(core1);
///     ENDWHILE();
///     
///     // outputs:
///     // -42
///     // 42
///  
/// # Related
/// * [pop]
void push(expr){
  expr = _wrapLiteral(expr);
  if(expr is _ExpressionNode){
    expr.compile(clobber : _setClobber);
    _setClobber = false;
    _HakkObject.addLine('push: __sys__.${expr.id}');
  }else{
    _HakkObject.addLine('push: ${expr}');    
  }
}

/// Pops a value off the stack.
/// 
/// See [push] for documentation of how push/pop work together.
///   
/// # Related
/// * [push]
/// 
/// @returns the value popped off the stack.
_ExpressionNode pop(){
  final node = new _ExpressionNode(val(0), val(0), 'pop');
  _expressionQueue.add(node);
  return node;
}

/// Returns the cosine of a numeric expression.
/// 
/// # Related
/// * [sine]
/// 
/// @param expr a numeric expression.
/// @returns the cosine of a numeric [expr] expression.
_ExpressionNode cosine(expr){
  expr = _wrapLiteral(expr);
  final node = new _ExpressionNode(expr, expr, 'cos');
  _expressionQueue.add(node);
  return node;
}


/// Returns the sine of a numeric expression.
/// 
/// # Related
/// * [cosine]
/// 
/// @param expr a numeric expression.
/// @returns the sine of a numeric [expr] expression.
_ExpressionNode sine(expr){
  expr = _wrapLiteral(expr);
  final node = new _ExpressionNode(expr, expr, 'sin');
  _expressionQueue.add(node);
  return node;
}

/// Calculates the result of a numeric [term] taken to a given [power].
/// 
/// Both parameters can themselves be expressions.
/// 
/// @param term is the numeric value to be operated on.
/// @param power is the exponent to apply to the term.
/// 
/// # Example
///      core0 << power(2, 4);
///      core0 << power(core0, 2);
_ExpressionNode power(term, power){
  term = _wrapLiteral(term);
  power = _wrapLiteral(power);

  final node = new _ExpressionNode(term, power, 'pow');
  _expressionQueue.add(node);
  return node;
}

/// Sets a memory [reference] to the result of a given [expr] expression.
/// 
/// Equivalent to the operator '<<' that is more commonly used.
/// 
/// # Example
///      set(core0, 42);
///      // equivalent to:
///      core0 << 42;
void set(_HakkObject reference, expr){
  expr = _wrapLiteral(expr);
  reference << expr;
}

/// Represents memory address 0 in the core block of a hakk.
final core0 = new _HakkObject.reference('core', 0);
/// Represents memory address 1 in the core block of a hakk.
final core1 = new _HakkObject.reference('core', 1);
/// Represents memory address 2 in the core block of a hakk.
final core2 = new _HakkObject.reference('core', 2);
/// Represents memory address 3 in the core block of a hakk.
final core3 = new _HakkObject.reference('core', 3);
/// Represents memory address 4 in the core block of a hakk.
final core4 = new _HakkObject.reference('core', 4);
/// Represents memory address 5 in the core block of a hakk.
final core5 = new _HakkObject.reference('core', 5);
/// Represents memory address 6 in the core block of a hakk.
final core6 = new _HakkObject.reference('core', 6);
/// Represents memory address 7 in the core block of a hakk.
final core7 = new _HakkObject.reference('core', 7);
/// Represents memory address 8 in the core block of a hakk.
final core8 = new _HakkObject.reference('core', 8);
/// Represents memory address 9 in the core block of a hakk.
final core9 = new _HakkObject.reference('core', 9);
/// Represents memory address 10 in the core block of a hakk.
final core10 = new _HakkObject.reference('core', 10);
/// Represents memory address 11 in the core block of a hakk.
final core11 = new _HakkObject.reference('core', 11);
/// Represents memory address 12 in the core block of a hakk.
final core12 = new _HakkObject.reference('core', 12);
/// Represents memory address 13 in the core block of a hakk.
final core13 = new _HakkObject.reference('core', 13);
/// Represents memory address 14 in the core block of a hakk.
final core14 = new _HakkObject.reference('core', 14);
/// Represents memory address 15 in the core block of a hakk.
final core15 = new _HakkObject.reference('core', 15);